from __future__ import annotations

import pytest
import stk


def build_kagome(
    lattice_size: tuple[int, int, int],
) -> stk.ConstructedMolecule:
    bb1 = stk.BuildingBlock("BrCCBr", [stk.BromoFactory()])
    bb2 = stk.BuildingBlock(
        smiles="BrC1C(Br)CC(Br)C(Br)C1",
        functional_groups=[stk.BromoFactory()],
    )
    cof = stk.ConstructedMolecule(
        topology_graph=stk.cof.Kagome(
            building_blocks=(bb1, bb2),
            lattice_size=lattice_size,
        ),
    )
    return cof


@pytest.fixture(
    params=(
        (1, 1, 1),
        (2, 2, 2),
        (4, 4, 4),
    ),
)
def lattice_size(request) -> tuple[int, int, int]:
    return request.param


def benchmark_kagome(
    benchmark,
    lattice_size: tuple[int, int, int],
) -> None:
    benchmark(build_kagome, lattice_size)



import itertools
import typing
from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator, Sequence

from stk._internal.building_block import BuildingBlock
from stk._internal.ea.crossover.record import CrossoverRecord
from stk._internal.ea.molecule_record import MoleculeRecord


class GeneticRecombination:
    """
    Recombine building blocks using biological systems as a model.

    Overall, this crosser mimics how animals and plants inherit
    DNA from their parents, except generalized to work with any
    number of parents. First it is worth discussing some
    terminology. A gene is a the smallest packet of genetic
    information. In animals, each gene can have multiple alleles.
    For example, there is a gene for hair color, and individual
    alleles for black, red, brown, etc. hair. This means that every
    person has a gene for hair color, but a person with black hair
    will have the black hair allele and a person with red hair will
    have the red hair allele. When two parents produce an
    offspring, the offspring will have a hair color gene and will
    inherit the allele of one of the parents at random. Therefore,
    if you have two parents, one with black hair and one with red
    hair, the offspring will either have black or red hair,
    depending on which allele they inherit.

    In an :mod:`.stk` :class:`.ConstructedMolecule`, each building
    block represents an allele. The question is, which gene is each
    building block an allele of? To answer that, let's first construct
    a couple of building block molecules

    .. testcode:: genetic-recombination

        import stk

        bb1 = stk.BuildingBlock(
            smiles='NCC(N)CN',
            functional_groups=[stk.PrimaryAminoFactory()],
        )
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        bb3 = stk.BuildingBlock(
            smiles='O=CCNC(C=O)C=O',
            functional_groups=[stk.AldehydeFactory()],
        )
        bb4 = stk.BuildingBlock(
            smiles='NCOCN',
            functional_groups=[stk.PrimaryAminoFactory()],
        )

    We can define a function which analyzes a building block
    molecule and returns the gene it belongs to, for example

    .. testcode:: genetic-recombination

        def get_gene(building_block):
            fg, = building_block.get_functional_groups(0)
            return type(fg)

    .. testcode:: genetic-recombination
        :hide:

        assert get_gene(bb1) == stk.PrimaryAmino
        assert get_gene(bb2) == stk.Aldehyde
        assert get_gene(bb3) == stk.Aldehyde
        assert get_gene(bb4) == stk.PrimaryAmino


    Here, we can see that the gene, to which each building block
    molecule belongs, is given by the class of its first functional
    group. Therefore there is an :class:`.PrimaryAmino` gene,
    which has two alleles ``bb1`` and ``bb4``, and there is an
    :class:`.Aldehyde` gene, which has two alleles ``bb2`` and ``bb3``.

    Alternatively, we could have defined a function such as

    .. testcode:: genetic-recombination

        def get_gene2(building_block):
            return building_block.get_num_functional_groups()

    .. testcode:: genetic-recombination
        :hide:

        assert get_gene2(bb1) == 3
        assert get_gene2(bb2) == 2
        assert get_gene2(bb3) == 3
        assert get_gene2(bb4) == 2

    Now we can see that we end up with the gene called
    ``3``, which has two alleles ``bb1`` and ``bb3``,
    and a second gene called ``2``, which has the alleles ``bb2`` and
    ``bb4``.

    To produce offspring molecules, this class categorizes
    each building block of the parent molecules into genes using
    the `get_gene` parameter. Then, to generate a single offspring, it
    picks a building block for every gene. The picked
    building blocks are used to construct the offspring. The
    topology graph of the offspring is one of the parent's.
    For obvious reasons, this approach works with any number of
    parents.

    Examples:
        *Crossing Constructed Molecules*

        Note that any number of parents can be used for the crossover

        .. testcode:: crossing-constructed-molecules

            import stk

            # Create the molecule records which will be crossed.

            bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
            bb2 = stk.BuildingBlock('O=CCCCC=O', [stk.AldehydeFactory()])
            graph1 = stk.polymer.Linear((bb1, bb2), 'AB', 2)
            polymer1  = stk.ConstructedMolecule(graph1)
            record1 = stk.MoleculeRecord(graph1)

            bb3 = stk.BuildingBlock('NCCCN', [stk.PrimaryAminoFactory()])
            bb4 = stk.BuildingBlock(
                smiles='O=C[Si]CCC=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            graph2 = stk.polymer.Linear((bb3, bb4), 'AB', 2)
            polymer2  = stk.ConstructedMolecule(graph2)
            record2 = stk.MoleculeRecord(graph2)

            # Create the crosser.

            def get_functional_group_type(building_block):
                fg, = building_block.get_functional_groups(0)
                return type(fg)

            recombination = stk.GeneticRecombination(
                get_gene=get_functional_group_type,
            )

            # Get the offspring molecules.

            cohort1 = tuple(recombination.cross(
                records=(record1, record2),
            ))

        .. testcode:: crossing-constructed-molecules
            :hide:

            _expected_cohort = (
                polymer1,
                polymer2,
                stk.ConstructedMolecule(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=(bb1, bb4),
                        repeating_unit='AB',
                        num_repeating_units=2,
                    ),
                ),
                stk.ConstructedMolecule(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=(bb2, bb3),
                        repeating_unit='AB',
                        num_repeating_units=2,
                    ),
                ),
            )

            def _get_smiles(item):
                if isinstance(item, stk.ConstructedMolecule):
                   return stk.Smiles().get_key(item)
                return stk.Smiles().get_key(
                    molecule=item.get_molecule_record().get_molecule(),
                 )

            _expected_smiles = set(map(_get_smiles, _expected_cohort))
            _cohort_smiles = set(map(_get_smiles, cohort1))
            assert _expected_smiles == _cohort_smiles

    """

    def __init__(
        self,
        get_gene: Callable[[BuildingBlock], typing.Any],
        name: str = "GeneticRecombination",
    ) -> None:
        """
        Parameters:
            get_gene:
                A function, which takes a :class:`.BuildingBlock`
                and returns its gene. To produce an offspring, one
                of the building blocks from each gene is picked.

            name:
                A name to identify the crosser instance.
        """

        self._get_gene = get_gene
        self._name = name

    def cross(
        self,
        records: Sequence[MoleculeRecord],
    ) -> Iterator[CrossoverRecord[MoleculeRecord]]:
        topology_graphs = (record.get_topology_graph() for record in records)
        for topology_graph, alleles in itertools.product(
            topology_graphs,
            self._get_alleles(records),
        ):

            def get_replacement(
                building_block: BuildingBlock,
            ) -> BuildingBlock:
                gene = self._get_gene(building_block)
                return next(
                    allele
                    for allele in alleles
                    if self._get_gene(allele) == gene
                )

            topology_graph = topology_graph.with_building_blocks(
                building_block_map={
                    building_block: get_replacement(building_block)
                    for building_block in topology_graph.get_building_blocks()
                },
            )
            yield CrossoverRecord(
                molecule_record=MoleculeRecord(
                    topology_graph=topology_graph,
                ),
                crosser_name=self._name,
            )

    def _get_alleles(
        self,
        records: Sequence[MoleculeRecord],
    ) -> Iterable[tuple[BuildingBlock, ...]]:
        """
        Yield every possible combination of alleles.

        """

        genes = defaultdict(list)
        topology_graphs = (record.get_topology_graph() for record in records)
        for topology_graph in topology_graphs:
            for allele in topology_graph.get_building_blocks():
                genes[self._get_gene(allele)].append(allele)
        return itertools.product(*genes.values())



import pytest
import stk

from ...case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cof.PeriodicLinkerlessHoneycomb(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1C2[C+]=NC2(Br)[C+](F)[C+](Br)" "[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    lattice_size=(2, 2, 1),
                ),
            ),
            smiles=(
                "[H]C12[C+]=NC13[C+](F)[C+]1[C+2][C+]2[C+]2[C+2][C+]4["
                "C+]5[C+2][C+]6[C+]7[C+2][C+]1[C+](F)C1(N=[C+]C71[H])C"
                "17N=[C+]C1([H])[C+]1[C+2][C+]([C+]8[C+2][C+]([C+]9[C+"
                "2][C+]([C+]%10[C+2][C+]1C1([H])[C+]=NC1([C+]%10F)C1(N"
                "=[C+]C61[H])[C+]5F)[C+](F)C1(N=[C+]C91[H])C1(N=[C+]C2"
                "1[H])[C+]4F)C1([H])[C+]=NC13[C+]8F)[C+]7F"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cof.PeriodicLinkerlessHoneycomb(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1C2[C+]=NC2(Br)[C+](F)[C+](Br)" "[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    lattice_size=(2, 2, 1),
                    optimizer=stk.PeriodicCollapser(),
                ),
            ),
            smiles=(
                "[H]C12[C+]=NC13[C+](F)[C+]1[C+2][C+]2[C+]2[C+2][C+]4["
                "C+]5[C+2][C+]6[C+]7[C+2][C+]1[C+](F)C1(N=[C+]C71[H])C"
                "17N=[C+]C1([H])[C+]1[C+2][C+]([C+]8[C+2][C+]([C+]9[C+"
                "2][C+]([C+]%10[C+2][C+]1C1([H])[C+]=NC1([C+]%10F)C1(N"
                "=[C+]C61[H])[C+]5F)[C+](F)C1(N=[C+]C91[H])C1(N=[C+]C2"
                "1[H])[C+]4F)C1([H])[C+]=NC13[C+]8F)[C+]7F"
            ),
            name=name,
        ),
    ),
)
def cof_periodic_linkerless_honeycomb(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import typing
from collections.abc import Iterator

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.selection.batch import Batch

from .selector import ExcludedBatches, IncludedBatches, Selector

T = typing.TypeVar("T", bound=MoleculeRecord)


class FilterMolecules(Selector[T]):
    """
    Allows a :class:`.Selector` to select only some molecules.

    Examples:

        *Using a Selection Algorithm on a Subset of Batches*

        .. testcode:: using-a-selection-algorithm-on-a-subset-of-batches

            import stk

            selector = stk.FilterMolecules(
                # Use only molecules in the top 5 batches of size 3.
                filter=stk.Best(num_batches=5, batch_size=3),
                # Select from those molecules using roulette selection.
                selector=stk.Roulette(num_batches=20, batch_size=3),
            )

            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): i
                for i in range(10)
            }

            for batch in selector.select(population):
                # Do stuff with batch. All the molecules in the batch
                # belong to the top 5 batches of size 3. The batch
                # was selected using roulette selection.
                pass
    """

    def __init__(self, filter: Selector[T], selector: Selector[T]) -> None:
        """
        Parameters:
            filter:
                Selects molecules which can be yielded by `selector`.

            selector:
                Selects batches of molecules. The batches can only
                contain molecules yielded by `filter`.
        """
        self._filter = filter
        self._selector = selector

    def select(
        self,
        population: dict[T, float],
        included_batches: "IncludedBatches" = None,
        excluded_batches: "ExcludedBatches" = None,
    ) -> Iterator[Batch[T]]:
        valid_batches = self._filter.select(
            population=population,
            included_batches=included_batches,
            excluded_batches=excluded_batches,
        )
        valid_population = {
            record: population[record]
            for batch in valid_batches
            for record in batch
        }
        yield from self._selector.select(
            population=valid_population,
            included_batches=included_batches,
            excluded_batches=excluded_batches,
        )



import numpy as np
import pytest
import stk


@pytest.fixture(
    scope="session",
    params=(
        lambda: stk.BuildingBlock("NCCN"),
        lambda: stk.BuildingBlock(
            smiles="Brc1ccc(Br)cc1Br",
            functional_groups=[stk.BromoFactory()],
        ),
        lambda: stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(
                    stk.BuildingBlock("BrCCBr", [stk.BromoFactory()]),
                    stk.BuildingBlock(
                        smiles="BrCNCCBr",
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
                repeating_unit="AB",
                num_repeating_units=2,
            ),
        ),
    ),
)
def molecule(request) -> stk.Molecule:
    """
    A :class:`.Molecule` instance with at least 2 atoms.


    """

    return request.param()


@pytest.fixture(
    params=[
        np.array([1.0, 0.0, 0.0]),
        np.array([0.0, 1.0, 0.0]),
        np.array([0.0, 0.0, 1.0]),
        np.array([1.0, 1.0, 1.0]),
    ],
)
def target(request):
    """
    The target vector onto which a molecule is rotated.

    """

    return request.param



"""
Terminal Alkene Factory
=======================

"""


from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_group_factories.utilities import get_atom_ids
from stk._internal.functional_groups.alkene import Alkene


class TerminalAlkeneFactory(FunctionalGroupFactory):
    # This docstring is a literal string, to silence a flake8 warning
    # about CH\ :sub:`2`. It's taking the backslash out of context.
    r"""
    Creates :class:`.Alkene` instances.

    Creates functional groups from substructures, which match the
    ``[*][C]([*])=[C]([H])[H]`` functional group string.

    Examples
    --------
    *Creating Functional Groups with the Factory*

    You want to create a building block which has
    :class:`.Alkene` functional groups, but only if they are terminal.
    You want the non-terminal carbon atom in those functional
    groups to be the *bonder* atom, and the terminal CH\ :sub:`2`
    group to be the *deleter* atoms.

    .. testcode:: creating-functional-groups-with-the-factory

        import stk

        building_block = stk.BuildingBlock(
            smiles='C=CCCCCC=C',
            functional_groups=(stk.TerminalAlkeneFactory(), ),
        )

    .. testcode:: creating-functional-groups-with-the-factory
        :hide:

        assert all(
            isinstance(functional_group, stk.Alkene)
            for functional_group
            in building_block.get_functional_groups()
        )
        assert building_block.get_num_functional_groups() == 2

    *Changing the Bonder and Deleter Atoms*

    You want to create a building block which has
    :class:`.Alkene` functional groups, but only if they are terminal.
    You want the carbon atoms to be the *bonder* atoms and you don't
    want any *deleter* atoms.

    .. testcode:: changing-the-bonder-and-deleter-atoms

        import stk

        terminal_alkene_factory = stk.TerminalAlkeneFactory(
            # The indices of the carbon atoms in the functional
            # group string (see docstring) are 1 and 3.
            bonders=(1, 3),
            deleters=(),
        )
        building_block = stk.BuildingBlock(
            smiles='C=CCCCCC=C',
            functional_groups=(terminal_alkene_factory, ),
        )

    .. testcode:: changing-the-bonder-and-deleter-atoms
        :hide:

        fg1, fg2 = building_block.get_functional_groups()
        assert fg1.get_num_bonders() == 2
        assert sum(1 for _ in fg1.get_deleters()) == 0
        assert fg2.get_num_bonders() == 2
        assert sum(1 for _ in fg2.get_deleters()) == 0

        assert all(
            isinstance(atom, stk.C)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_bonders()
        )

    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(
        self,
        bonders=(1,),
        deleters=(3, 4, 5),
        placers=None,
    ):
        """
        Initialize a :class:`.TerminalAlkeneFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        ids = get_atom_ids("[*][C]([*])=[C]([H])[H]", molecule)
        for atom_ids in ids:
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield Alkene(
                carbon1=atoms[1],
                atom1=atoms[0],
                atom2=atoms[2],
                carbon2=atoms[3],
                atom3=atoms[4],
                atom4=atoms[5],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



import pytest
import stk

from ....case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.OnePlusOne(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1[C+2][C+](Br)[C+]2[C+](F)[C+2]"
                                "C2(Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                ),
            ),
            smiles=(
                "F[C+]1[C+2]C23[C+2][C+]4[C+2][C+]([C+]12)C12[C+2][C+]"
                "(F)[C+]1[C+]3[C+2][C+]4[C+2]2"
            ),
            name=name,
        ),
    ),
)
def cage_one_plus_one(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import pytest
import stk

from ...case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cof.Square(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)C(F)(Br)[C+]1Br",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    lattice_size=(2, 2, 1),
                ),
            ),
            smiles=(
                "FC1(Br)C2=C(C3=C([C+]=N3)C3(F)C(=C(C4=C(Br)[C+]=N4)[C"
                "+]3Br)C3=C([C+]=N3)[C+]3C(C4=C(Br)[C+]=N4)=C(C4=C(Br)"
                "[C+]=N4)C3(F)C3=C(N=[C+]3)C3=C(C4=C(Br)[C+]=N4)C(F)(B"
                "r)[C+]3C3=C2N=[C+]3)[C+]1Br"
            ),
            name=name,
        ),
    ),
)
def cof_square(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import typing
from collections.abc import Callable, Iterator, Sequence

import numpy as np

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.selection.batch import Batch
from stk._internal.ea.selection.selectors.yielded_batches import YieldedBatches
from stk._internal.key_makers.inchi import Inchi
from stk._internal.key_makers.molecule import MoleculeKeyMaker

from .selector import Selector

T = typing.TypeVar("T", bound=MoleculeRecord)


class StochasticUniversalSampling(Selector[T]):
    """
    Yields batches of molecules through stochastic universal sampling.

    Stochastic universal sampling lays out batches along a line, with
    each batch taking up length proportional to its fitness. It
    then creates a set of evenly spaced pointers to different points
    on the line, each of which is occupied by a batch. Batches which
    are pointed to are yielded.

    This approach means weaker members of the population
    are given a greater chance to be chosen than in
    :class:`.Roulette` selection [#]_.

    References:

    .. [#] https://en.wikipedia.org/wiki/Stochastic_universal_sampling

    Examples:

        *Yielding Single Molecule Batches*

        Yielding molecules one at a time. For example, if molecules need
        to be selected for mutation or the next generation.

        .. testcode:: yielding-single-molecule-batches

            import stk

            # Make the selector.
            stochastic_sampling = stk.StochasticUniversalSampling(5)

            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): i
                for i in range(100)
            }

            # Select the molecules.
            for selected, in stochastic_sampling.select(population):
                # Do stuff with each selected molecule.
                pass
    """

    def __init__(
        self,
        num_batches: int | None = None,
        batch_size: int = 1,
        duplicate_molecules: bool = True,
        duplicate_batches: bool = True,
        key_maker: MoleculeKeyMaker = Inchi(),
        fitness_modifier: Callable[
            [dict[T, float]], dict[T, float]
        ] = lambda x: x,
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """
        Parameters:

            num_batches:
                The number of batches to yield. If ``None`` then yielding
                will continue forever or until the generator is exhausted,
                whichever comes first.

            batch_size:
                The number of molecules yielded at once.

            duplicate_molecules:
                If ``True`` the same molecule can be yielded in more than
                one batch.

            duplicate_batches:
                If ``True`` the same batch can be yielded more than once.

            key_maker:
                Used to get the keys of molecules. If two molecules have
                the same key, they are considered duplicates.

            fitness_modifier:
                Takes the `population` on which :meth:`~.Selector.select`
                is called and returns a :class:`dict`, which maps records
                in the `population` to the fitness values the
                :class:`.Selector` should use.

            random_seed:
                The random seed to use.
        """
        super().__init__(key_maker, fitness_modifier, batch_size)

        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)

        self._generator = random_seed
        self._duplicate_molecules = duplicate_molecules
        self._duplicate_batches = duplicate_batches
        self._num_batches = (
            float("inf") if num_batches is None else num_batches
        )

    def _select_from_batches(
        self,
        batches: Sequence[Batch[T]],
        yielded_batches: YieldedBatches[T],
    ) -> Iterator[Batch[T]]:
        batches = sorted(batches, reverse=True)

        # SUS may need to run multiple rounds if duplicate_molecules or
        # duplicate_batches is False. This is because in each round
        # you can generate multiple pointers to the same batch or to
        # batches sharing molecules. If this happens the lower fitness
        # batch will not be yielded. Instead a second round of SUS will
        # occur with any ineligible batches removed and a reduced
        # number of pointers, to account for batches yielded in the
        # previous rounds. This will repeat until the desired number
        # of batches has been yielded, or there are no more valid
        # batches.
        while batches and yielded_batches.get_num() < self._num_batches:
            yield from self._select_with_stochastic_universal_sampling(
                batches=batches,
                yielded_batches=yielded_batches,
            )

            if yielded_batches.get_num() < self._num_batches:
                if not self._duplicate_molecules:
                    batches_ = filter(
                        yielded_batches.has_no_yielded_molecules,
                        batches,
                    )
                if not self._duplicate_batches:
                    batches_ = filter(
                        yielded_batches.is_unyielded_batch,
                        batches,
                    )
                if (
                    not self._duplicate_molecules
                    or not self._duplicate_batches
                ):
                    batches = tuple(batches_)

    def _select_with_stochastic_universal_sampling(
        self,
        batches: Sequence[Batch[T]],
        yielded_batches: YieldedBatches[T],
    ) -> Iterator[Batch[T]]:
        total = sum(batch.get_fitness_value() for batch in batches)
        batch_positions = []
        batch_position = 0.0
        for batch in batches:
            batch_position += batch.get_fitness_value() / total
            batch_positions.append(batch_position)

        num_batches = typing.cast(
            int,
            min(self._num_batches - yielded_batches.get_num(), len(batches)),
        )
        pointer_distance = 1 / num_batches
        pointers = []
        pointer = self._generator.uniform(0, pointer_distance)
        for _ in range(num_batches):
            pointers.append(pointer)
            pointer += pointer_distance

        batch_index = 0
        for pointer in pointers:
            while pointer > batch_positions[batch_index]:
                batch_index += 1

            batch = batches[batch_index]

            if (
                not self._duplicate_molecules
                and yielded_batches.has_yielded_molecules(batch)
            ):
                continue

            if (
                not self._duplicate_batches
                and yielded_batches.is_yielded_batch(batch)
            ):
                continue

            yield batch



"""
Functional Group Sorter
=======================

"""

import numpy as np

from stk._internal.utilities.utilities import get_acute_vector

from .sorter import _Sorter


class _FunctionalGroupSorter(_Sorter):
    """
    Sorts functional groups according to their angle.

    """

    __slots__ = [
        "_items",
        "_reference",
        "_axis",
        "_placer_centroid",
        "_building_block",
    ]

    def __init__(self, building_block):
        """
        Initialize a :class:`._FunctionalGroupSorter` instance.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block, whose functional groups are to be
            sorted.

        """

        self._building_block = building_block
        fg0_position = building_block.get_centroid(
            atom_ids=next(
                building_block.get_functional_groups()
            ).get_placer_ids(),
        )
        self._placer_centroid = placer_centroid = building_block.get_centroid(
            atom_ids=building_block.get_placer_ids(),
        )
        fg0_direction = fg0_position - placer_centroid
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        axis = np.cross(
            fg0_direction,
            get_acute_vector(
                reference=core_centroid - placer_centroid,
                vector=building_block.get_plane_normal(),
            ),
        )
        axis.setflags(write=False)
        super().__init__(
            items=range(building_block.get_num_functional_groups()),
            reference=fg0_direction,
            axis=axis,
        )

    def _get_vector(self, item):
        building_block = self._building_block
        (fg,) = building_block.get_functional_groups(item)
        fg_position = building_block.get_centroid(fg.get_placer_ids())
        return fg_position - self._placer_centroid



"""
Bromo
=====

"""

from __future__ import annotations

import typing

from stk._internal.atom import Atom
from stk._internal.elements import Br

from .generic_functional_group import GenericFunctionalGroup


class Bromo(GenericFunctionalGroup):
    """
    Represents a bromo functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[bromine][atom]``.

    """

    def __init__(
        self,
        bromine: Br,
        atom: Atom,
        bonders: tuple[Atom, ...],
        deleters: tuple[Atom, ...],
        placers: typing.Optional[tuple[Atom, ...]] = None,
    ) -> None:
        """
        Initialize a :class:`.Bromo` instance.

        Parameters:

            bromine:
                The ``[bromine]`` atom.

            atom:
                The ``[atom]`` atom.

            bonders:
                The bonder atoms.

            deleters:
                The deleter atoms.

            placers:
                The placer atoms. If ``None`` the `bonders` will be
                used.

        """

        self._bromine = bromine
        self._atom = atom
        super().__init__(
            atoms=(bromine, atom),
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_bromine(self) -> Br:
        """
        Get the ``[bromine]`` atom.

        Returns:

            The ``[bromine]`` atom.

        """

        return self._bromine

    def get_atom(self) -> Atom:
        """
        Get the ``[atom]`` atom.

        Returns:

            The ``[atom]`` atom.

        """

        return self._atom

    def clone(self) -> Bromo:
        clone = super()._clone()
        clone._bromine = self._bromine
        clone._atom = self._atom
        return clone

    def _with_ids(
        self,
        id_map: dict[int, int],
    ) -> Bromo:
        super()._with_ids(id_map)
        if (bromine_id := self._bromine.get_id()) in id_map:
            self._bromine = self._bromine.with_id(id_map[bromine_id])

        if (atom_id := self._atom.get_id()) in id_map:
            self._atom = self._atom.with_id(id_map[atom_id])

        return self

    def with_ids(
        self,
        id_map: dict[int, int],
    ) -> Bromo:
        return self.clone()._with_ids(id_map)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{self._bromine}, {self._atom}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



"""
Sorter
======

"""

import numpy as np

from stk._internal.utilities.utilities import vector_angle


class _Sorter:
    """
    Sorts items according to their angle from a reference vector.

    """

    __slots__ = ["_items", "_reference", "_axis"]

    def __init__(self, items, reference, axis):
        """
        Initialize a :class:`._Sorter`.

        Parameters
        ----------
        items : :class:`iterable` of :class:`object`
            The items to be sorted.

        reference : :class:`numpy.ndarray`
            The reference from which the angle is calculated.

        axis : :class:`numpy.ndarray`
            A vector orthogonal to `reference`, used to determine
            which direction is clockwise. Must be an immutable
            array.

        """

        self._items = items
        self._reference = reference
        self._axis = axis

    def _get_vector(self, item):
        """
        Get the vector according to which `item` should be sorted.

        Parameters
        ----------
        item : :class:`object`
            The item being sorted.

        Returns
        -------
        :class:`numpy.ndarray`
            The vector of the `item`, which should be used to measure
            its angle with respect to the *reference*.

        """

        raise NotImplementedError()

    def _get_angle(self, item):
        """
        Get the angle of `vector` relative to `reference`.

        Parameters
        ----------
        item : :class:`object`
            The item being sorted.

        Returns
        -------
        :class:`float`
            The angle between `item` and the reference vector.

        """

        vector = self._get_vector(item)
        theta = vector_angle(self._reference, vector)
        projection = vector @ self._axis
        if theta > 0 and projection < 0:
            return 2 * np.pi - theta
        return theta

    def get_items(self):
        """
        Yield the sorted items.

        Yields
        ------
        :class:`object`
            An item.

        """

        yield from sorted(self._items, key=self._get_angle)

    def get_axis(self):
        """
        Get the axis used to determine which direction is clockwise.

        Returns
        -------
        :class:`numpy.ndarray`
            The axis. The array is immutable.

        """

        return self._axis



"""
M12L24
======

"""

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import LinearVertex, NonLinearVertex


class M12L24(Cage):
    """
    Represents a cage topology graph.

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(4)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=NC=CC(C2=CC=CC(C3=C'
                'C=NC=C3)=C2)=C1'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M12L24(
                building_blocks=(bb1, bb2),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    Metal building blocks with four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 4-functional groups: 0 to 11
        | 2-functional groups: 12 to 35

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        NonLinearVertex(0, [1, 0, 0]),
        NonLinearVertex(1, [-1, 0, 0]),
        NonLinearVertex(2, [0, 1, 0]),
        NonLinearVertex(3, [0, -1, 0]),
        NonLinearVertex(4, [0.5, 0.5, 0.707]),
        NonLinearVertex(5, [0.5, -0.5, 0.707]),
        NonLinearVertex(6, [-0.5, 0.5, 0.707]),
        NonLinearVertex(7, [-0.5, -0.5, 0.707]),
        NonLinearVertex(8, [0.5, 0.5, -0.707]),
        NonLinearVertex(9, [0.5, -0.5, -0.707]),
        NonLinearVertex(10, [-0.5, 0.5, -0.707]),
        NonLinearVertex(11, [-0.5, -0.5, -0.707]),
        LinearVertex(12, [0.9, 0.31, 0.31], False),
        LinearVertex(13, [0.9, 0.31, -0.31], False),
        LinearVertex(14, [0.9, -0.31, 0.31], False),
        LinearVertex(15, [0.9, -0.31, -0.31], False),
        LinearVertex(16, [-0.9, 0.31, 0.31], False),
        LinearVertex(17, [-0.9, 0.31, -0.31], False),
        LinearVertex(18, [-0.9, -0.31, 0.31], False),
        LinearVertex(19, [-0.9, -0.31, -0.31], False),
        LinearVertex(20, [0.31, 0.9, 0.31], False),
        LinearVertex(21, [0.31, 0.9, -0.31], False),
        LinearVertex(22, [-0.31, 0.9, 0.31], False),
        LinearVertex(23, [-0.31, 0.9, -0.31], False),
        LinearVertex(24, [0.31, -0.9, 0.31], False),
        LinearVertex(25, [0.31, -0.9, -0.31], False),
        LinearVertex(26, [-0.31, -0.9, 0.31], False),
        LinearVertex(27, [-0.31, -0.9, -0.31], False),
        LinearVertex(28, [0.58, 0, 0.82], False),
        LinearVertex(29, [-0.58, 0, 0.82], False),
        LinearVertex(30, [0, 0.58, 0.82], False),
        LinearVertex(31, [0, -0.58, 0.82], False),
        LinearVertex(32, [0.58, 0, -0.82], False),
        LinearVertex(33, [-0.58, 0, -0.82], False),
        LinearVertex(34, [0, 0.58, -0.82], False),
        LinearVertex(35, [0, -0.58, -0.82], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[12]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[13]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[14]),
        Edge(3, _vertex_prototypes[0], _vertex_prototypes[15]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[16]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[17]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[18]),
        Edge(7, _vertex_prototypes[1], _vertex_prototypes[19]),
        Edge(8, _vertex_prototypes[2], _vertex_prototypes[20]),
        Edge(9, _vertex_prototypes[2], _vertex_prototypes[21]),
        Edge(10, _vertex_prototypes[2], _vertex_prototypes[22]),
        Edge(11, _vertex_prototypes[2], _vertex_prototypes[23]),
        Edge(12, _vertex_prototypes[3], _vertex_prototypes[24]),
        Edge(13, _vertex_prototypes[3], _vertex_prototypes[25]),
        Edge(14, _vertex_prototypes[3], _vertex_prototypes[26]),
        Edge(15, _vertex_prototypes[3], _vertex_prototypes[27]),
        Edge(16, _vertex_prototypes[4], _vertex_prototypes[28]),
        Edge(17, _vertex_prototypes[4], _vertex_prototypes[30]),
        Edge(18, _vertex_prototypes[4], _vertex_prototypes[12]),
        Edge(19, _vertex_prototypes[4], _vertex_prototypes[20]),
        Edge(20, _vertex_prototypes[5], _vertex_prototypes[14]),
        Edge(21, _vertex_prototypes[5], _vertex_prototypes[24]),
        Edge(22, _vertex_prototypes[5], _vertex_prototypes[28]),
        Edge(23, _vertex_prototypes[5], _vertex_prototypes[31]),
        Edge(24, _vertex_prototypes[6], _vertex_prototypes[16]),
        Edge(25, _vertex_prototypes[6], _vertex_prototypes[29]),
        Edge(26, _vertex_prototypes[6], _vertex_prototypes[30]),
        Edge(27, _vertex_prototypes[6], _vertex_prototypes[22]),
        Edge(28, _vertex_prototypes[7], _vertex_prototypes[18]),
        Edge(29, _vertex_prototypes[7], _vertex_prototypes[26]),
        Edge(30, _vertex_prototypes[7], _vertex_prototypes[31]),
        Edge(31, _vertex_prototypes[7], _vertex_prototypes[29]),
        Edge(32, _vertex_prototypes[8], _vertex_prototypes[13]),
        Edge(33, _vertex_prototypes[8], _vertex_prototypes[32]),
        Edge(34, _vertex_prototypes[8], _vertex_prototypes[34]),
        Edge(35, _vertex_prototypes[8], _vertex_prototypes[21]),
        Edge(36, _vertex_prototypes[9], _vertex_prototypes[15]),
        Edge(37, _vertex_prototypes[9], _vertex_prototypes[32]),
        Edge(38, _vertex_prototypes[9], _vertex_prototypes[35]),
        Edge(39, _vertex_prototypes[9], _vertex_prototypes[25]),
        Edge(40, _vertex_prototypes[10], _vertex_prototypes[17]),
        Edge(41, _vertex_prototypes[10], _vertex_prototypes[23]),
        Edge(42, _vertex_prototypes[10], _vertex_prototypes[34]),
        Edge(43, _vertex_prototypes[10], _vertex_prototypes[33]),
        Edge(44, _vertex_prototypes[11], _vertex_prototypes[19]),
        Edge(45, _vertex_prototypes[11], _vertex_prototypes[33]),
        Edge(46, _vertex_prototypes[11], _vertex_prototypes[27]),
        Edge(47, _vertex_prototypes[11], _vertex_prototypes[35]),
    )

    _num_windows = 14
    _num_window_types = 2



from .utilities import are_equivalent_bonds


def test_get_deleted_bonds(case_data):
    """
    Test that the correct bonds are deleted by a :class:`.Reaction`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the reaction to test and the bonds which
        should be deleted.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_deleted_bonds(
        reaction_result=case_data.reaction.get_result(),
        deleted_bonds=case_data.deleted_bonds,
    )


def _test_get_deleted_bonds(reaction_result, deleted_bonds):
    """
    Test that the correct bonds are deleted by `reaction_result`.

    Parameters
    ----------
    reaction_result : :class:`.ReactionResult`
        The result of a reaction.

    new_bonds : :class:`tuple` of :class:`.Bond`
        The bonds, which should be deleted by `reaction_result`.

    Returns
    -------
    None : :class:`NoneType`

    """

    are_equivalent_bonds(
        reaction_result.get_deleted_bonds(),
        deleted_bonds,
    )



import pytest
import stk
from pytest_lazyfixture import lazy_fixture

from ...case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=[C+]1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="AB",
                    num_repeating_units=4,
                ),
            ),
            smiles=(
                "[C+]1=[C+]C2=C1C1=C(N=[C+]1)C1=C([C+]=[C+]1)C1=C(N=[C"
                "+]1)C1=C([C+]=[C+]1)C1=C(N=[C+]1)C1=C([C+]=[C+]1)C1=C"
                "2N=[C+]1"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrN1N(Br)N=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="A",
                    num_repeating_units=8,
                ),
            ),
            smiles=(
                "N1=NN2N1N1N=NN1N1N=NN1N1N=NN1N1N=NN1N1N=NN1N1N=NN1N1N" "=NN21"
            ),
            name=name,
        ),
    ),
)
def macrocycle_macrocycle(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )


@pytest.fixture(
    params=(lazy_fixture("macrocycle_macrocycle"),),
)
def macrocycle(request):
    return request.param



import stk

from .utilities import is_equivalent_atom


def test_repr(bond):
    """
    Test :meth:`.Bond.__repr__`.

    Parameters
    ----------
    bond : :class:`.Bond`
        The bond whose representation is tested.

    Returns
    -------
    None : :class:`NoneType`

    """

    other = eval(repr(bond), dict(stk.__dict__))
    is_equivalent_atom(other.get_atom1(), bond.get_atom1())
    is_equivalent_atom(other.get_atom2(), bond.get_atom2())
    assert other.get_order() == bond.get_order()
    assert other.get_periodicity() == bond.get_periodicity()



"""
Null Optimizer
==============

"""

from .optimizer import Optimizer


class NullOptimizer(Optimizer):
    """
    Does not perform an optimization.

    """

    def optimize(self, state):
        return state



"""
Optimizer Utilities
===================

This module defines utilities for optimizers.

"""

from collections import defaultdict

import mchammer as mch


def get_mch_bonds(state):
    """
    Yield the bonds of the :mod:`MCHammer` molecule.

    Parameters
    ----------
    state : :class:`.ConstructionState`
        The state of the molecule under construction.

    Yields
    ------
    :class:`MCHammer.Bond`
        A bond in the molecule.

    """

    for i, bond_infos in enumerate(state.get_bond_infos()):
        ba1 = bond_infos.get_bond().get_atom1().get_id()
        ba2 = bond_infos.get_bond().get_atom2().get_id()
        # Must ensure bond atom id ordering is the same here as in
        # line 30. Therefore, sort here.
        ba1, ba2 = sorted((ba1, ba2))
        yield mch.Bond(id=i, atom_ids=(ba1, ba2))


def get_long_bond_ids(state):
    """
    Yield the ids of the long bonds to optimize.

    Parameters
    ----------
    state : :class:`.ConstructionState`
        The state of the molecule under construction.

    Yields
    ------
    :class:`tuple` of :class:`int`
        A pair of atom ids that identify a bond to be optimized.

    """

    for i, bond_infos in enumerate(state.get_bond_infos()):
        ba1 = bond_infos.get_bond().get_atom1().get_id()
        ba2 = bond_infos.get_bond().get_atom2().get_id()
        # None if for constructed bonds.
        if bond_infos.get_building_block() is None:
            yield sorted((ba1, ba2))


def get_subunits(state):
    """
    Get connected graphs based on building block ids.

    Parameters
    ----------
    state : :class:`.ConstructionState`
        The state of the molecule under construction.

    Returns
    -------
    :class:`.dict`
        The subunits of the molecule split by building block id. Key is
        subunit identifier, Value is :class:`list` of atom ids in
        subunit.

    """

    subunits = defaultdict(list)
    for atom_info in state.get_atom_infos():
        subunits[atom_info.get_building_block_id()].append(
            atom_info.get_atom().get_id()
        )

    return subunits



from scipy.spatial.distance import euclidean


def get_maximum_diameter(position_matrix, atom_ids):
    atomic_positions = position_matrix[atom_ids, :]
    return float(
        euclidean(
            atomic_positions.min(axis=0),
            atomic_positions.max(axis=0),
        )
    )



"""
A utility for dropping the databases created by running the tests.

This script is not part of the test suite. It is meant to be run
manually by developers when they feel the need to.

"""


import pymongo


def main():
    databases_to_drop = (
        "_stk_pytest_database",
        "_stk_test_database_for_testing",
    )
    client = pymongo.MongoClient()
    for database in databases_to_drop:
        client.drop_database(database)


if __name__ == "__main__":
    main()



from .utilities import is_equivalent_atom


def test_with_id(atom, id):
    """
    Test :meth:`.Atom.with_id`.

    Parameters
    ----------
    atom : :class:`.Atom`
        The atom to test.

    id : :class:`int`
        The correct id of the new atom.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save a clone to ensure that "atom" is not changed by the test.
    before = atom.clone()
    _test_with_id(atom, id)
    is_equivalent_atom(before, atom)


def _test_with_id(atom, id):
    """
    Test :meth:`.Atom.with_id`.

    Parameters
    ----------
    atom : :class:`.Atom`
        The atom to test.

    id : :class:`int`
        The correct id of the new atom.

    Returns
    -------
    None : :class:`NoneType`

    """

    new_atom = atom.with_id(id)
    assert new_atom is not atom
    assert new_atom.__class__ is atom.__class__
    assert new_atom.get_id() == id
    assert new_atom.get_charge() == atom.get_charge()
    assert new_atom.get_atomic_number() == atom.get_atomic_number()



import pathlib
import typing

import stk

from .case_data import CaseData

T = typing.TypeVar("T", bound=stk.MoleculeRecord)


def test_selection_plotter(
    tmp_path: pathlib.Path,
    case_data: CaseData,
) -> None:
    _test_selection_plotter(
        selector=case_data.selector,
        population=case_data.population,
        filename=tmp_path / "selection",
    )


def _test_selection_plotter(
    selector: stk.Selector[T],
    population: dict[T, float],
    filename: pathlib.Path,
) -> None:
    stk.SelectionPlotter(filename, selector)
    tuple(selector.select(population))
